using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using maxstAR;
using System;
using UnityEngine.EventSystems;

public class ARPointerTracker : ARBehaviour
{
    private CameraBackgroundBehaviour cameraBackgroundBehaviour = null;
    private int screenWidth;
    private int screenHeight;
    private int cameraWidth = 1280;
    private int cameraHeight = 720;
    private Matrix4x4 currentPose = Matrix4x4.identity;
    private int anchorCount = 0;
    private List<string> anchorList = new List<string>();
    private Dictionary<String, Matrix4x4> anchorDictionary = new Dictionary<String, Matrix4x4>();
    private Dictionary<String, GameObject> anchorGameObjectDictionary = new Dictionary<String, GameObject>();
    public GameObject trackingObject;
    public GameObject arCamera;
    public GameObject pointObject;
    String anchorName = "";
    // Start is called before the first frame update

    void Awake()
    {

        Init();

        AndroidRuntimePermissions.Permission[] result = AndroidRuntimePermissions.RequestPermissions("android.permission.WRITE_EXTERNAL_STORAGE", "android.permission.CAMERA");
        if (result[0] == AndroidRuntimePermissions.Permission.Granted && result[1] == AndroidRuntimePermissions.Permission.Granted)
            Debug.Log("We have all the permissions!");
        else
            Debug.Log("Some permission(s) are not granted...");

        cameraBackgroundBehaviour = FindObjectOfType<CameraBackgroundBehaviour>();
        if (cameraBackgroundBehaviour == null)
        {
            Debug.LogError("Can't find CameraBackgroundBehaviour.");
            return;
        }
    }

    void Start()
    {
        screenWidth = Screen.width;
        screenHeight = Screen.height;

        CameraDevice.GetInstance().Start();
        TrackerManager.GetInstance().StartTracker(0x1000);
        StartCoroutine(FindSurface());
    }

    void OnDestroy()
    {
        CameraDevice.GetInstance().Stop();
        TrackerManager.GetInstance().StopTracker();
        TrackerManager.GetInstance().DestroyTracker();
    }

    void OnApplicationPause(bool pause)
    {
        if (pause)
        {
            TrackerManager.GetInstance().StopTracker();
        }
        else
        {
            CameraDevice.GetInstance().Start();
            TrackerManager.GetInstance().StartTracker(0x1000);
            StartCoroutine(FindSurface());
        }
    }

    IEnumerator FindSurface()
    {
        yield return new WaitForSeconds(0.2f);
        anchorList.Clear();
        TrackerManager.GetInstance().QuitFindingSurface();
        TrackerManager.GetInstance().FindSurface();
    }

    // Update is called once per frame
    void Update()
    {
        if(screenWidth != Screen.width)
        {
            screenWidth = Screen.width;
            screenHeight = Screen.height;
        }

        TrackingState state = TrackerManager.GetInstance().UpdateTrackingState();

        if (state == null)
        {
            return;
        }

        cameraBackgroundBehaviour.UpdateCameraBackgroundImage(state);
        cameraWidth = state.GetImage().GetWidth();
        cameraHeight = state.GetImage().GetHeight();
        

        TrackingResult trackingResult = state.GetTrackingResult();

        int trackingCount = trackingResult.GetCount();
        if (trackingCount == 0)
        {
            return;
        }

        if (Input.GetMouseButtonUp(0))
        {
            UpdateTouchDelta(Input.mousePosition);
        }

        for (int i = 0; i < trackingCount; i++)
        {
            Trackable trackable = trackingResult.GetTrackable(i);

            if (trackable.GetName() == "status")
            {
                continue;
            }
            else if (trackable.GetName() == "default")
            {
                currentPose = trackable.GetOriginalPose();
                continue;
            }

            if (!anchorDictionary.ContainsKey(trackable.GetName()))
            {
                anchorDictionary.Add(trackable.GetName(), trackable.GetOriginalPose());
                RemoveAnchor(trackable.GetName());
            }
        }

        foreach (string anchorName in anchorDictionary.Keys)
        {
            Matrix4x4 anchorPose = anchorDictionary[anchorName];
            Matrix4x4 currentAnchorPose = anchorPose;
            Matrix4x4 multiplyPose = currentPose * currentAnchorPose;
            Matrix4x4 targetPose = MatrixUtils.GetConvertPose(multiplyPose);

            GameObject anchorGameObject = anchorGameObjectDictionary[anchorName];

            anchorGameObject.transform.position = MatrixUtils.PositionFromMatrix(targetPose);
            anchorGameObject.transform.rotation = MatrixUtils.QuaternionFromMatrix(targetPose);
            anchorGameObject.transform.localScale = MatrixUtils.ScaleFromMatrix(targetPose);
        }
    }

    private String GetAnchorName(bool next)
    {
        if (next)
            anchorCount++;

        return "anchor" + anchorCount;
    }
    private void UpdateTouchDelta(Vector2 touchPosition)
    {
        if(Input.GetMouseButtonUp(0))
        {
            PointerEventData pointer = new PointerEventData(EventSystem.current);
            List<RaycastResult> raycastResult = new List<RaycastResult>();

            pointer.position = touchPosition;
            EventSystem.current.RaycastAll(pointer, raycastResult);

            foreach (RaycastResult result in raycastResult)
            {
                if (result.gameObject.layer == 5)
                {
                    return;
                }
            }

            float end_x = touchPosition.x;
            float end_y = screenHeight - touchPosition.y;
            Matrix4x4 lastPose = currentPose;

            anchorName = GetAnchorName(true);
            Vector3 end_ncoord = ScreenToLocal(end_x, end_y);
            AddPrevAnchor(anchorName, end_ncoord.x, end_ncoord.y, MatrixToFloat(lastPose));

            GameObject cloneObject = GameObject.Instantiate(pointObject);
            anchorGameObjectDictionary[anchorName] = cloneObject;
            cloneObject.transform.parent = trackingObject.transform;
            cloneObject.transform.localScale = new Vector3(0.2f, 0.2f, 0.2f);
            cloneObject.SetActive(true);
        }
    }

    private Vector3 ScreenToLocal(float screenX, float screenY)
    {
        float sw = screenWidth;
        float sh = screenHeight;

        float cw = cameraWidth;
        float ch = cameraHeight;

        float nw = (sw > sh) ? sw : sh;
        float nh = (sw > sh) ? sh : sw;

        float px = (sw > sh) ? screenX : screenY;
        float py = (sw > sh) ? screenY : nh - screenX;

        float rw = nw / cw;
        float rh = nh / ch;

        float margin_nh = 0.0f;
        float margin_nw = 0.0f;

        // cut height
        if (rw > rh)
        {
            float vh = rw * ch;
            margin_nh = (vh - nh) * 0.5f;
            py += margin_nh;
        }
        else
        {
            float vw = rh * cw;
            margin_nw = (vw - nw) * 0.5f;
            px += margin_nw;
        }

        float nx = (px - (nw * 0.5f + margin_nw)) / nw;
        float ny = (py - (nh * 0.5f + margin_nh)) / nw;

        return new Vector3(nx, ny, 1.0f);
    }

    void AddPrevAnchor(String anchorName, float normalizedCenterPosX, float normalizedCenterPosY, float[] prevPose)
    {
        String strPose = "";
        for (int i = 0; i < 15; i++)
            strPose += prevPose[i] + ",";
        strPose += prevPose[15];
        String addCommand = "{\"vivar\":\"add_prev_anchor_normalized\",\"name\":\"" + anchorName + "\",\"x_pos\":" + normalizedCenterPosX + ",\"y_pos\":" + normalizedCenterPosY +
                ",\"search_pose\":\"" + strPose + "\"}";

        TrackerManager.GetInstance().AddTrackerData(addCommand, false);

        anchorList.Add(anchorName);

        Debug.Log(addCommand);
    }

    void RemoveAnchor(String anchorName)
    {
        String removeCommand = "{\"vivar\":\"remove_anchor\",\"name\":\"" + anchorName + "\"}";
        TrackerManager.GetInstance().AddTrackerData(removeCommand, false);
    }

    private float[] MatrixToFloat(Matrix4x4 matrix)
    {
        float[] returnMatrix = new float[16];
        returnMatrix[0] = matrix.m00;
        returnMatrix[1] = matrix.m10;
        returnMatrix[2] = matrix.m20;
        returnMatrix[3] = matrix.m30;

        returnMatrix[4] = matrix.m01;
        returnMatrix[5] = matrix.m11;
        returnMatrix[6] = matrix.m21;
        returnMatrix[7] = matrix.m31;

        returnMatrix[8] = matrix.m02;
        returnMatrix[9] = matrix.m12;
        returnMatrix[10] = matrix.m22;
        returnMatrix[11] = matrix.m32;

        returnMatrix[12] = matrix.m03;
        returnMatrix[13] = matrix.m13;
        returnMatrix[14] = matrix.m23;
        returnMatrix[15] = matrix.m33;

        return returnMatrix;
    }
}
